//+------------------------------------------------------------------+
//|                                                Master-Copier.mq4 |
//|                                   Copyright 2023, Just Up Coding |
//|                                            https://www.justup.us |
//+------------------------------------------------------------------+
#property copyright "Copyright 2023, Just Up Coding"
#property link      "https://www.justup.us"
#property version   "1.00"
#property strict

int openAmount = 0;
int openTickets[];

//+------------------------------------------------------------------+
//| Expert initialization function                                   |-------------------------------------------------------------------------------------------|
//+------------------------------------------------------------------+
int OnInit()
  {
//---
   //Test();

   //DeleteAllClassObjects();
   InitArray();
//---
   return(INIT_SUCCEEDED);
  }
//+------------------------------------------------------------------+
//| Expert deinitialization function                                 |-------------------------------------------------------------------------------------------|
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
//---
   DeleteAllClassObjects();
  }
//+------------------------------------------------------------------+
//| Expert tick function                                             |-------------------------------------------------------------------------------------------|
//+------------------------------------------------------------------+
void OnTick()
  {
//---
      Main();
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OnChartEvent(const int id,const long& lparam,const double& dparam,const string& sparam)
  {

  }
//+------------------------------------------------------------------+
//+ Classes and Structures                                           |-------------------------------------------------------------------------------------------|
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CSignalsFile
  {
private:
   int               orderNumber;
   string            name;

public:
                     CSignalsFile(void);
   void              Write(string signal);
   string            LastReportedStatus();
  };

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CSignalsFile::CSignalsFile(void)
  {
   this.orderNumber = 0;

   while(!FileIsExist("signalNo_" + (string)this.orderNumber + ".signal",FILE_COMMON))
     {
      this.orderNumber++;
     }


   this.name = "signalNo_" + (string)orderNumber + ".signal";

  }

//+------------------------------------------------------------------+
//| Writes a signal in the file                                      |
//+------------------------------------------------------------------+
void CSignalsFile::Write(string signal)
  {
   int handle;

   handle = FileOpen(this.name,FILE_COMMON|FILE_TXT|FILE_WRITE);
   FileWrite(handle,signal);
   FileClose(handle);
  }

//+------------------------------------------------------------------+
//| Read the first parameter of the signal                           |
//+------------------------------------------------------------------+
string CSignalsFile::LastReportedStatus(void)
  {
   int handle;
   string str;
   string strSignals[];


   handle = FileOpen(this.name,FILE_COMMON|FILE_TXT|FILE_READ|FILE_SHARE_READ);
   str = FileReadString(handle);

   StringSplit(str,44,strSignals);

   str = strSignals[0];

   return str;
  }
//+------------------------------------------------------------------+
//| This class handles the open positions                            |
//+------------------------------------------------------------------+
class COpenPositions
  {
private:
   int               ticket;  //Ticket of the order
   string            symbol;  //Symbol of the order
   string            type;    //Type of the order from ENUM_ORDER_TYPE converted to string
   double            lots;    //The lot size
   double            price;   //The limit price, in case of a pending position
   double            sl;      //The stop loss price
   double            tp;      //The take profits price
   bool              UpdateValues(bool newOrder);   //Update the values of the class
   bool              stillActive;   //Still active the signals
   CSignalsFile      sFile;   //The file handler


public:
   bool              CheckForUpdates(bool newOrder);  //Check for an update of a current order
   void              SetTicket(int extTicket) {this.ticket = extTicket; this.CheckForUpdates(true); this.stillActive = true;} //Set the ticket
   int               GetTicket() {return this.ticket;} //Get the current signal ticket
  };
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
bool COpenPositions::CheckForUpdates(bool newOrder)
  {
   Print("TEST");
   
   if(this.ticket == 0)
      return false;

   if(!this.stillActive)
      return false;

   if(newOrder)
     {
      this.UpdateValues(true);
      return true;
     }

   if(this.sFile.LastReportedStatus() == "ORDER_MISSED")
     {
      this.stillActive = false;
      return false;
     }

   if(OrderSelect(this.ticket,SELECT_BY_TICKET))
     {
      //--- From Main Pool
      if(OrderCloseTime() == 0)
        {
            //--- Limit price changed
            if(OrderType() != OP_BUY && OrderType() != OP_SELL && OrderOpenPrice() != this.price)
               { UpdateValues(false); return true; }
            //--- SL Price changed
            if(OrderStopLoss() != this.sl)
               { UpdateValues(false); return true; }
            //--- TP Price changed
            if(OrderTakeProfit() != this.tp)
               { UpdateValues(false); return true; }
            if(OrderLots() != this.lots)
               { UpdateValues(false); return true; }
        }

      //--- From History Pool
      else
        {
         
            //--- Check if the close/delete signal was received and copied
            if(this.sFile.LastReportedStatus() == "ORDER_CLOSED")
               this.stillActive = false;
            else
               this.sFile.Write("CLOSE_OR_DELETE");

            return true;
        }
     }

   return false;
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool COpenPositions::UpdateValues(bool newOrder)
  {
   if(!OrderSelect(this.ticket, SELECT_BY_TICKET))
      return false;

   //--- If it is a new order
   if(newOrder)
     {
      this.symbol = OrderSymbol();
      if(OrderType() == OP_BUY || OrderType() == OP_SELL)
         this.price = -1;
      else
         this.price = OrderOpenPrice();
      this.sl = OrderStopLoss();
      this.tp = OrderTakeProfit();
      this.lots = OrderLots();

      switch(OrderType())
        {
         case  OP_BUY:
            this.type = "MARKET_BUY";
            break;
         case  OP_SELL:
            this.type = "MARKET_SELL";
            break;
         case  OP_BUYLIMIT:
            this.type = "LIMIT_BUY";
            break;
         case  OP_SELLLIMIT:
            this.type = "LIMIT_SELL";
            break;
         case  OP_BUYSTOP:
            this.type = "STOP_BUY";
            break;
         case  OP_SELLSTOP:
            this.type = "STOP_SELL";
            break;

         default:
            this.type = "INVALID_TRADE";
            break;
        }
        
        //--- Sent Signal
        this.sFile.Write(
                           "NEW_ORDER," + 
                           this.type + "," +
                           this.symbol + "," +
                           DoubleToStr(this.price,5) + "," +
                           DoubleToStr(this.sl,5) + "," + 
                           DoubleToStr(this.tp,5) + "," + 
                           DoubleToStr(this.lots,2)
                        );
     }
   //--- Check for the order change
   else
     {
         
     }
   return false;
  }
//+------------------------------------------------------------------+

COpenPositions *signals[];

//+------------------------------------------------------------------+
//+ Functions                                                        |-------------------------------------------------------------------------------------------|
//+------------------------------------------------------------------+


//+------------------------------------------------------------------+
//| Initialize the array                                             |
//+------------------------------------------------------------------+
void InitArray()
  {
   openAmount = OrdersTotal();

   ArrayResize(openTickets,openAmount);
   ArrayResize(signals,openAmount);

   for(int i=0; i<openAmount; i++)
     {
      if(!OrderSelect(i,SELECT_BY_POS,MODE_TRADES))
         Print("An error ocurred while selecting the order | " + __FUNCTION__);

      openTickets[i] = OrderTicket();
      signals[ArraySize(signals) - 1] = new COpenPositions;
      signals[ArraySize(signals) - 1].SetTicket(OrderTicket());
     }

   openAmount = OrdersTotal();
  }



//+------------------------------------------------------------------+
//| Check for the opening/close of an order                          |
//+------------------------------------------------------------------+
void CheckForOrdersChange()
  {
//--- IF a new position has been detected
   if(OrdersTotal() > openAmount) //--- New Position
     {
      int tempTicket = 0;

      for(int i=0; i<OrdersTotal(); i++)
        {
         if(!OrderSelect(i,SELECT_BY_POS))
            Print("An error ocurred while selecting the order | " + __FUNCTION__);

         if(ArraySize(openTickets) == 0)
           {
            tempTicket = OrderTicket();
            break;
           }

         for(int x=0; x<ArraySize(openTickets); x++)
           {
            if(OrderTicket() == openTickets[x])
               break;
            else
               tempTicket = OrderTicket();
           }
        }
      
      if(tempTicket == 0)
         return;

      ArrayResize(openTickets,ArraySize(openTickets) + 1);
      openTickets[ArraySize(openTickets) - 1] = tempTicket;
      openAmount = ArraySize(openTickets);
      
      Print("Aquiii");

      //--- SIGNALS - NEW
      ArrayResize(signals,openAmount);
      Print("Aquiii222");
      signals[ArraySize(signals)] = new COpenPositions;
      Print("Aquiii333");
      signals[ArraySize(signals)].SetTicket(tempTicket);
      Print("Aquiii444");
     }

//--- IF a position close/delete has been detected
   if(OrdersTotal() < openAmount)
     {

      int ticketPos = -1;

      for(int i=0; i<ArraySize(openTickets); i++)
        {
         if(!OrderSelect(OrdersHistoryTotal() - 1,SELECT_BY_POS,MODE_HISTORY))
            Print("An error ocurred while selecting the order | " + __FUNCTION__);

         if(OrderTicket() == openTickets[i])
           {
            ticketPos = i;
            break;
           }
        }

      if(ticketPos == -1)
         return;

      int tempArray[];
      ArrayResize(tempArray, OrdersTotal());
      int controller = 0;

      for(int i=0; i<openAmount; i++)
        {

         if(i == ticketPos)
            continue;

         tempArray[controller] = openTickets[i];

        }

      ArrayResize(openTickets,0);
      ArrayResize(openTickets,OrdersTotal());
      ArrayCopy(openTickets,tempArray,0,0,WHOLE_ARRAY);

      openAmount = OrdersTotal();
     }
  }


//+------------------------------------------------------------------+
//| Main function to be called from the OnTick() function            |
//+------------------------------------------------------------------+
void Main()
  {
     Print("Checking...");
      CheckForOrdersChange();
   
      Print(ArraySize(signals));
      for(int i=1; i<=ArraySize(signals); i++)
      {
         signals[i].CheckForUpdates(false);
      }
  }

//+------------------------------------------------------------------+
//| Deletes all the class objects                                    |
//+------------------------------------------------------------------+
void DeleteAllClassObjects()
  {
      for(int i=0;i<ArraySize(signals);i++)
        {
            delete signals[i];
        }
  }
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Temporal testing function                                        |
//+------------------------------------------------------------------+
int Test()
  {
   string x = ",";
   uchar xS[];

   StringToCharArray(x,xS);

   Print(xS[0]);

   return INIT_SUCCEEDED;
  }
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//+ Posible signals                                                  |-------------------------------------------------------------------------------------------|
//+------------------------------------------------------------------+
/*
   ORDER_CLOSED - The slave has received the close or delete signal and has been executed.
   CLOSE_OR_DELETE - The master has sent a close or delete signal and waiting for the slave receive confirmation.
   NEW_ORDER - The master has sent a new order signal and waiting for the salve receive confirmation
   NEW_ORDER_RECEIVED - The slave has successfully received and executed the new order signal
   ORDER_MODOFY - The master has sent a modification of one of the parameters
   ORDER_MODIFICATION_RECEIVED - The slave has receiived and executed the trade modification
   ORDER_MISSED - This is a signal gived by the slave that the order was executed on the master but missed by the slave and the master has sent the first modification. So will be ignored by the master.
*/
//+------------------------------------------------------------------+

 